# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####
import bpy
from .helpers_group import is_inst_coll
from mathutils import  Vector, Euler

# these input names seems shared for both radial and linear arrays
use_object = 'Input_11'
inst_input = 'Input_2'
cnt_input = 'Input_3'

input_names = {
    'GP_RadialArray': {
        'object': 'Input_2',
        'count': 'Input_3',
        'scale': 'Input_8',
        'rotation': 'Input_10',
        'circle_normal': 'Input_9'
    },
    'GP_LinearArray': {
        'object': 'Input_2',
        'count': 'Input_3',
        'scale': 'Input_9',
        'rotation': 'Input_10',
        'offset_mode': 'Input_8', # bool
    }
}

def get_first_array_mod(obj):
    for mod in obj.modifiers:
        if mod.type == 'NODES' and mod.node_group and mod.node_group.name in {'GP_RadialArray', 'GP_LinearArray'}:
            return mod
    return None

def get_geonodes_mod_type(obj, node_group_name):
    for mod in obj.modifiers:
        if mod.type == 'NODES' and mod.node_group and mod.node_group.name == node_group_name:
            return mod
    return None



# ['Input_10'] - rotation, ['Input_8'] - scale > for both radial and linear arrays
# when editing group, apply neg scale, and rotation to fix jumping of instances inside array modifiers
def apply_mw_inv_to_array_mods(edited_coll, empty_mw):
    ''' Should fix jumping of instances inside array modifiers when editing group'''
    # empty_loc, empty_rot, empty_scale = matrix_decompose(empty_mw) # decomposed in matrix form
    loc, rotQuat, scale = empty_mw.decompose()

    scale = Vector((1/scale[0], 1/scale[1], 1/scale[2]))
    rotQuat.invert()

    # find all obj that use same empty.instance_collection
    for obj in bpy.data.objects:
        if obj.type == 'MESH' and len(obj.data.vertices)==0: # arry holder obj has no vertices
            array_mod = get_first_array_mod(obj)
            if array_mod:
                if array_mod[use_object] and array_mod[inst_input]:
                    arr_ins = array_mod[inst_input]
                    if is_inst_coll(arr_ins) and arr_ins.instance_collection == edited_coll:
                        # apply neg scale and rot to obj
                        in_names = input_names[array_mod.node_group.name]
                        scale_in_name = in_names['scale']
                        inst_scale = Vector(array_mod[scale_in_name][:]) # Scale for rot, but Offset for linear
                        rot_in_name = in_names['rotation']
                        inst_rot_euler = Euler(array_mod[rot_in_name][:], 'XYZ')

                        # apply  tran
                        inst_scale = inst_scale * scale
                        inst_rot_euler.rotate(rotQuat)

                        # write back to array mod
                        array_mod['Input_8'] = inst_scale[:]
                        array_mod['Input_10'] = inst_rot_euler[:]
                obj.update_tag()
    bpy.context.view_layer.update()


# same thing but in reverse (for closing group)
def apply_mw_to_array_mods(edited_coll, empty_mw):
    ''' Should fix jumping of instances inside array modifiers when editing group'''
    # empty_loc, empty_rot, empty_scale = matrix_decompose(empty_mw) # decomposed in matrix form
    loc, rotQuat, scale = empty_mw.decompose()

    # find all obj that use same empty.instance_collection
    for obj in bpy.data.objects:
        if obj.type == 'MESH' and len(obj.data.vertices)==0: # arry holder obj has no vertices
            array_mod = get_first_array_mod(obj)
            if array_mod:
                if array_mod[use_object] and array_mod[inst_input]:
                    arr_ins = array_mod[inst_input]
                    if is_inst_coll(arr_ins) and arr_ins.instance_collection == edited_coll:
                        # apply neg scale and rot to obj
                        inst_scale = Vector(array_mod['Input_8'][:])
                        inst_rot_euler = Euler(array_mod['Input_10'][:], 'XYZ')

                        # apply  tran
                        inst_scale = inst_scale * scale
                        inst_rot_euler.rotate(rotQuat)

                        # write back to array mod
                        array_mod['Input_8'] = inst_scale[:]
                        array_mod['Input_10'] = inst_rot_euler[:]
                obj.update_tag()
    bpy.context.view_layer.update()
