import bpy
from .helpers_group import localView_up


class EditedGroupsSettings(bpy.types.PropertyGroup):
    currentEmptyName: bpy.props.StringProperty()
    was_library: bpy.props.BoolProperty(default=False)
    lib_path: bpy.props.StringProperty(default='', subtype='FILE_PATH')

class CollectionsSettings(bpy.types.PropertyGroup):
    bbox_obj: bpy.props.PointerProperty(name='Bbox Obj', type=bpy.types.Object)
    to_parent: bpy.props.PointerProperty(name="Original link", type=bpy.types.Collection)
    to_proxy: bpy.props.PointerProperty(name="Proxy link", type=bpy.types.Collection)
    is_proxy: bpy.props.BoolProperty(name='is_proxy', default=False)



class GPSceneSettings(bpy.types.PropertyGroup):
    bbox_enabled: bpy.props.BoolProperty(name='enable snapping bbox', default=False, options={'HIDDEN'})
    viewcolor_override_enabled: bpy.props.BoolProperty(name='view color override', default=False, options={'HIDDEN'})
    groupLocalView: bpy.props.BoolProperty( name="Isolate edited group", description="Hide all objects outside currently edited group", default=False, update=localView_up)
    groupLocalViewDepth: bpy.props.IntProperty(name="Depth", description="Maximum hierarchy visibility depth (default=1)", default=1, min=1, max=5, subtype='UNSIGNED', step=1, update=localView_up)
    use_stored_origin: bpy.props.BoolProperty(name='Use stored Flip Origin', description='Use Stored flip origin (if exist) position rather than current cursor position', default=True)


class FlipOpSettings(bpy.types.PropertyGroup):
    origin: bpy.props.EnumProperty(name='Origin', description='Origin', default='WORLD',
                                   items=[('WORLD', 'World', '', '', 0),
                                          ('CURSOR', '3D Cursor', '', '', 1)])
    flip_coordinates: bpy.props.EnumProperty(name="Orientation", description="Coordinates System for mirroring", default="GLOBAL",
                                             items=(("GLOBAL", "GLOBAL", ""),
                                                    ("LOCAL", "LOCAL", ""),
                                                    ("CURSOR", "CURSOR", "")
                                                    ))
    flip_axis: bpy.props.EnumProperty(name="Flip Axis", description="",  options={'ENUM_FLAG'},  # default={"X"},
                                      items=(("X", "X", "", 2),
                                             ("Y", "Y", "", 4),
                                             ("Z", "Z", "", 8)))
    cursor_mat: bpy.props.FloatVectorProperty(size=16, subtype="MATRIX")


class GPObjectSettings(bpy.types.PropertyGroup):
    gpro_hide: bpy.props.BoolProperty(name="Hide object", default=False)
    flip_props: bpy.props.PointerProperty(type=FlipOpSettings)
    orig_viewport_color: bpy.props.FloatVectorProperty(name="Viewport Color", subtype='COLOR', size=4, default=[1.0, 1.0, 1.0, 1.0], min=0.0, max=1.0)
    # mirror_matrix_pivot_inv: bpy.props.FloatVectorProperty(size=16)

def group_from_coll(self, context):
    layout = self.layout
    layout.operator('object.gp_from_collection')


def register_props():
    bpy.types.Scene.storedGroupSettings = bpy.props.CollectionProperty(type=EditedGroupsSettings)
    bpy.types.Scene.gp_props = bpy.props.PointerProperty(type=GPSceneSettings)

    bpy.types.Collection.BBoxVertsList = bpy.props.FloatVectorProperty(size=24)  # minx, maxx, miny, maxy, minz, maxz
    bpy.types.Collection.gp_props = bpy.props.PointerProperty(type=CollectionsSettings)
    bpy.types.Collection.created_with_gp = bpy.props.BoolProperty(default=False, name='Created with GroupPro')

    bpy.types.Object.gp_props = bpy.props.PointerProperty(type=GPObjectSettings)
    bpy.types.OUTLINER_MT_collection.append(group_from_coll)


def unregister_props():
    del bpy.types.Scene.storedGroupSettings
    del bpy.types.Scene.gp_props

    del bpy.types.Collection.gp_props
    del bpy.types.Collection.BBoxVertsList
    del bpy.types.Collection.created_with_gp
    del bpy.types.Object.gp_props

    bpy.types.OUTLINER_MT_collection.remove(group_from_coll)
