# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####
import bpy

from .group_pro import *

#########            UI PANELS        ###################################


class GP_PT_GroupProPanel(bpy.types.Panel):
    bl_idname = "GP_PT_GroupProPanel"
    bl_label = "GroupPro"
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'UI'
    bl_category = 'GroupPro'
    bl_context = 'objectmode'
    # bl_options = {'HIDE_HEADER'}

    def draw(self, context):
        layout = self.layout.box()
        col = layout.column(align=True)
        row = col.row(align=True)
        split = row.split(factor=0.70, align=True)
        split.prop(context.scene.gp_props, 'groupLocalView', icon='RESTRICT_VIEW_OFF')
        split.prop(context.scene.gp_props, 'groupLocalViewDepth', slider=False, text='')
        create = col.operator(GP_OT_GroupProCreate.bl_idname)

        if context.active_object:
            create.name = context.active_object.name

        row = col.row(align=True)
        row.operator(GP_OT_GroupProEdit.bl_idname)
        row.operator(GP_OT_GroupProRayCastOpen.bl_idname, icon='EYEDROPPER', text='')
        if len(bpy.context.scene.storedGroupSettings) > 0:
            row = col.row(align=True)
            row.operator(GP_OT_GroupProAdd.bl_idname)
            row.operator(GP_OT_GroupProRemove.bl_idname)
            col.operator(GP_OT_GroupProClose.bl_idname)


        if context.active_object is not None and context.active_object.type == "EMPTY" and context.active_object.instance_collection:
            col.menu(GP_MT_SetOrigin.bl_idname, text="Set Group origin")

        box = layout.box()
        addon_prefs = get_addon_preferences()
        icon = "TRIA_DOWN" if addon_prefs.expandPanel else "TRIA_RIGHT"
        box.prop(addon_prefs, "expandPanel", icon=icon,
                 text="Additional operations:", emboss=False)
        if addon_prefs.expandPanel:
            sub_coll = box.column(align=True)
            sub_coll.menu('GP_MT_mat_sub_menu')
            geoOper = sub_coll.operator(GP_OT_GroupProDissolve.bl_idname, icon='GROUP_VERTEX')
            geoOper.maxDepth = 0
            sub_coll.operator(GP_OT_GroupProFlip.bl_idname, icon='AUTOMERGE_ON')
            ic = 'PINNED' if context.scene.gp_props.use_stored_origin else 'UNPINNED'
            draw_pin = context.active_object is not None and (context.active_object.gp_props.flip_props.origin == 'CURSOR')
            if draw_pin:
                row.prop(context.scene.gp_props, 'use_stored_origin', icon=ic, icon_only=True)
            oper = sub_coll.operator(GP_OT_GroupProMakeUnique.bl_idname, icon='UNLINKED')
            oper.maxDepth = 1
            sub_coll.operator(GP_OT_ApplyLocRotScale.bl_idname, icon='EMPTY_ARROWS')
            sub_coll.operator(GP_OT_GroupProInitialize.bl_idname, icon='FILE_REFRESH')
            sub_coll.operator(GP_OT_GroupProDelete.bl_idname, icon='PANEL_CLOSE')
            sub_coll.operator(GP_OT_GroupProCleanUp.bl_idname, icon='TRASH')
            sub_coll.operator("wm.call_menu_pie", text="Proxy", icon='GHOST_DISABLED').name = GP_MT_GProPieProxy.bl_idname


class GP_MT_GProPieOrigin(bpy.types.Menu):
    bl_idname = "GP_MT_grouppro_origin_pie"
    bl_label = "GroupPro Origin Pie"

    def draw(self, context):
        layout = self.layout
        pie = layout.menu_pie()

        op = pie.operator(GP_OT_GroupProSetOrginCenter.bl_idname, text="3D cursor", icon='PIVOT_CURSOR')
        op.method = '3DCURSOR'
        op = pie.operator(GP_OT_GroupProSetOrginCenter.bl_idname, text="Objects Center", icon='CENTER_ONLY')
        op.method = 'CENTER'
        op = pie.operator(GP_OT_GroupProSetOrginCenter.bl_idname, text="Bounding box bottom", icon='ALIGN_BOTTOM')
        op.method = 'BBB'
        op = pie.operator(GP_OT_GroupProSetOrginCenter.bl_idname, text="Bounding box center", icon='LIGHTPROBE_CUBEMAP')
        op.method = 'BBC'


class GP_MT_GProPieProxy(bpy.types.Menu):
    bl_idname = "GP_MT_grouppro_proxy_pie"
    bl_label = "GroupPro Proxy"

    def draw(self, context):
        layout = self.layout
        pie = layout.menu_pie()

        pie.operator("object.gp_proxy_create", icon='GHOST_DISABLED')
        pie.operator("object.gp_proxy_switch", icon='FILE_REFRESH')
        pie.operator("object.gp_delete_proxy", icon='TRASH')


class GP_MT_GProPie(bpy.types.Menu):
    bl_idname = "GP_MT_grouppro_main_pie"
    bl_label = "GroupPro Pie"

    def draw(self, context):
        layout = self.layout
        pie = layout.menu_pie()

        # 4 - LEFT
        if context.active_object is not None and context.active_object.type == "EMPTY" and context.active_object.instance_collection:
            pie.operator("wm.call_menu_pie", text="Set Origin").name = GP_MT_GProPieOrigin.bl_idname
        else:
            pie.separator()

        # 6 - RIGHT
        pie.operator(GP_OT_GroupProClose.bl_idname, icon='STICKY_UVS_LOC')

        # 2 - BOTTOM
        split = pie.split()
        col = split.column(align=True)
        row = col.row(align=True)
        row.scale_y = 1.5
        geoOper = row.operator(GP_OT_GroupProDissolve.bl_idname, icon='GROUP_VERTEX')
        geoOper.maxDepth = 0

        row.scale_y = 1.5
        row.menu('GP_MT_mat_sub_menu')

        row = col.row(align=True)
        row.scale_y = 1.5
        row.operator(GP_OT_GroupProFlip.bl_idname, icon='AUTOMERGE_ON')
        ic = 'PINNED' if context.scene.gp_props.use_stored_origin else 'UNPINNED'
        draw_pin = context.active_object is not None and (context.active_object.gp_props.flip_props.origin == 'CURSOR')
        if draw_pin:
            row.prop(context.scene.gp_props, 'use_stored_origin', icon=ic, icon_only=True)

        row = col.row(align=True)
        row.scale_y = 1.5
        oper = row.operator(GP_OT_GroupProMakeUnique.bl_idname, icon='UNLINKED')

        row = col.row(align=True)
        row.scale_y = 1.5

        row.scale_y = 1.5
        row.menu('GP_MT_modifiers_menu')


        row = col.row(align=True)
        row.scale_y = 1.5
        row.operator(GP_OT_ApplyLocRotScale.bl_idname, icon='EMPTY_ARROWS')

        row = col.row(align=True)
        row.scale_y = 1.5
        row.operator(GP_OT_GroupProDelete.bl_idname, icon='PANEL_CLOSE')

        # oper.maxDepth = 1 #! not used much - moved to panel
        # row = col.row(align=True)
        # row.scale_y = 1.5
        # row.operator(GP_OT_GroupProInitialize.bl_idname, icon='FILE_REFRESH')

        # row = col.row(align=True) #! not used much - moved to panel
        # row.scale_y = 1.5
        # row.operator(GP_OT_GroupProCleanUp.bl_idname, icon='TRASH')

        row = col.row(align=True)
        row.scale_y = 1.5
        row.operator("wm.call_menu_pie", text="Proxy", icon='GHOST_DISABLED').name = GP_MT_GProPieProxy.bl_idname
        # row.operator(GP_OT_CreateProxy.bl_idname, icon='GHOST_DISABLED')


        # row.template_list("MATERIAL_UL_matslots", "", context.object, "material_slots", context.object, "active_material_index", rows=4)
        # 8 - TOP
        create = pie.operator(GP_OT_GroupProCreate.bl_idname, icon='STICKY_UVS_LOC')
        if context.active_object:
            create.name = context.active_object.name

        # 7 - TOP - LEFT
        pie.operator(GP_OT_GroupProRayCastOpen.bl_idname, icon='EYEDROPPER')

        # 9 - TOP - RIGHT
        pie.operator(GP_OT_GroupProEdit.bl_idname, icon='TRIA_DOWN')

        # 1 - BOTTOM - LEFT
        if len(bpy.context.scene.storedGroupSettings) > 0:
            split = pie.split(factor=0.7)
            row = split.row(align=True)
            row.scale_y = 1.5
            row.prop(context.scene.gp_props, 'groupLocalView', text='Isolate', icon='RESTRICT_VIEW_OFF')
            spl2 = split.split()
            row = spl2.row(align=True)
            row.scale_y = 1.5
            row.prop(context.scene.gp_props, 'groupLocalViewDepth', slider=False, text='')

        else:
            pie.separator()

        # 3 - BOTTOM - RIGHT
        if len(bpy.context.scene.storedGroupSettings) > 0:
            split = pie.split()
            col = split.column(align=True)
            row = col.row(align=True)
            row.scale_y = 1.5
            row.operator(GP_OT_GroupProAdd.bl_idname, icon='ADD')
            row = col.row(align=True)
            row.scale_y = 1.5
            row.operator(GP_OT_GroupProRemove.bl_idname, icon='REMOVE')
        else:
            pie.separator()

class VIEW3D_PT_Color(bpy.types.Panel): #puover for material list menu
    bl_parent_id = "GP_PT_GroupProPanel"
    bl_idname = 'VIEW3D_PT_Color'
    bl_label = 'Change Color'
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'UI'
    bl_category = 'GroupPro'
    bl_context = 'objectmode'

    def draw(self, context):
        layout = self.layout
        if hasattr(context, 'parent_mat'):
            col = layout.column(align=True)
            col.prop(context.parent_mat, 'diffuse_color') #passed by context override
            col.prop(context.parent_mat, 'metallic') #passed by context override
            col.prop(context.parent_mat, 'roughness') #passed by context override
            if context.parent_mat.node_tree:
                if 'Principled BSDF' in context.parent_mat.node_tree.nodes.keys():
                    principled = context.parent_mat.node_tree.nodes['Principled BSDF'].inputs['Base Color']
                    col = layout.column(align=True)
                    col.prop(principled, 'default_value', text='Principled Diffuse Color')


class GP_MT_MaterialListMenu(bpy.types.Menu):
    bl_idname = "GP_MT_sub_material_list_menu"
    bl_label = "Material Scene list"

    ''' Material Menu without empty slot Material fill (fill_empty_slots = False) '''
    def draw(self, context):
        layout = self.layout

        #* Second entry:  update color
        parent_mat_name = context.parent_mat.name if hasattr(context, 'parent_mat') and context.parent_mat else ''
        for mat in bpy.data.materials:
            #* With use same mat - just change color
            if mat.name == parent_mat_name:
                # col.prop(mat, 'diffuse_color', text='Update Color', icon='COLOR')
                layout.popover(panel='VIEW3D_PT_Color')
                # op = col.operator("object.gpro_assign_mat", text='Update Color', icon='COLOR') #! TOO Damm slow!
                # op.parent_mat = parent_mat_name
                # op.replacement_mat = mat.name  # replacement_mat == mat -> replace color
                # op.new_mat = False
                # op.mat_color = mat.diffuse_color[:]
                # row.prop(mat, 'diffuse_color')
                break

        layout.label(text='Replace with:')
        layout.separator()
        col = layout.column(align=True)

        #* First entry: new Mat
        op = col.operator("object.gpro_assign_mat", text='New material', icon='MATERIAL')
        op.parent_mat = parent_mat_name  # if null, set empty slots to new mat
        op.replacement_mat = ''
        op.new_mat = True
        #op.mat_color #only for new mat or updating mat.diffuse_color

        if bpy.data.materials:
            for mat in bpy.data.materials:
                if mat.name != parent_mat_name:  # * Replace from existing mat
                    op = col.operator("object.gpro_assign_mat", text=mat.name, icon_value=layout.icon(mat))
                    op.parent_mat = parent_mat_name
                    op.replacement_mat = mat.name
                    op.new_mat = False
                    #op.mat_color #only for new mat or updating mat.diffuse_color
        else:
            layout.label(text="No materials")

class GP_MT_MaterialOverrideMenu(bpy.types.Menu):
    bl_idname = "GP_MT_sub_material_override_menu"
    bl_label = "Material Scene list"

    ''' Material Menu without empty slot Material fill (fill_empty_slots = False) '''
    def draw(self, context):
        layout = self.layout
        layout.label(text='Replace with:')
        layout.separator()
        col = layout.column(align=True)

        #* First entry: new Mat
        op = col.operator("object.gpro_force_mat", text='New material', icon='MATERIAL')
        op.replacement_mat = ''
        op.new_mat = True
        #op.mat_color #only for new mat or updating mat.diffuse_color

        if bpy.data.materials:
            for mat in bpy.data.materials:
                op = col.operator("object.gpro_force_mat", text=mat.name, icon_value=layout.icon(mat))
                op.replacement_mat = mat.name
                op.new_mat = False
        else:
            layout.label(text="No materials")



class GP_MT_MaterialCollectionMenu(bpy.types.Menu):
    bl_idname = "GP_MT_mat_sub_menu"
    bl_label = "Materials"

    ''' Draw menu with materials list, empty slot. Each with submenu with list of materials, to replace it '''
    empty_mat_exist = False
    def get_obj_materials(self, obj):
        mat_list = set()
        if obj.type not in {'MESH', 'CURVE'}:
            return mat_list
        if not obj.material_slots:
            self.empty_mat_exist = True
        for slot in obj.material_slots:
            if slot.material:
                mat_list.add(slot.material.name)
            else:
                self.empty_mat_exist = True
        return mat_list

    checked_collections = []

    def get_mat_rec(self, current_group):
        mat_list = set()
        if current_group.name in self.checked_collections: #skip if already checked
            return set()
        for obj in current_group.objects:
            if obj == current_group.gp_props.bbox_obj:  # do not process bbox mesh
                continue
            if obj.instance_collection and obj.type == 'EMPTY':
                mat_list |= self.get_mat_rec(obj.instance_collection)
            else:
                mat_list |= self.get_obj_materials(obj)

        #add sub colls support
        for child_coll in current_group.children:
            sub_objs = child_coll.all_objects
            for obj in sub_objs:
                if obj.type == 'EMPTY' and obj.instance_collection:
                    mat_list |= self.get_mat_rec(obj.instance_collection)
                else:
                    mat_list |= self.get_obj_materials(obj)
            self.checked_collections.append(child_coll.name)
        self.checked_collections.append(current_group.name)
        return mat_list


    def draw(self, context):
        self.checked_collections.clear()
        layout = self.layout
        col = layout.column(align=True)
        mat_names_from_coll = set()
        for obj in context.selected_objects:
            if obj.type == 'EMPTY' and obj.instance_collection:  # get mats recursive from coll
                mat_names_from_coll |= self.get_mat_rec(obj.instance_collection)
            else:
                mat_names_from_coll |= self.get_obj_materials(obj)

        if mat_names_from_coll or self.empty_mat_exist: #replace materials
            if self.empty_mat_exist:  # call modified sub_empty_material_list_menu
                col.context_pointer_set("parent_mat", None)  # done: can wa pass value to Menu?? With This workaround yes
                col.menu('GP_MT_sub_material_list_menu', text='Empty Slots')
                col.menu('GP_MT_sub_material_override_menu', text='Override All')
                col.separator()
            for mat_name in mat_names_from_coll:  # DO replace mats
                mat = bpy.data.materials[mat_name]
                col.context_pointer_set("parent_mat", mat)  # done: can wa pass value to Menu?? With This workaround yes
                col.menu('GP_MT_sub_material_list_menu', text=mat_name, icon_value=layout.icon(mat))

        elif not context.selected_objects:
            layout.label(text="No Selection")
        else:
            layout.label(text="No materials")

# linear array and radial array menu
class GP_MT_ModifiersMenu(bpy.types.Menu):
    bl_idname = "GP_MT_modifiers_menu"
    bl_label = "Modifiers"

    def draw(self, context):
        layout = self.layout
        layout.operator('group_pro.edit_array', icon='MODIFIER')
        layout.operator('group_pro.linear_array', text='Linear Array', icon='MOD_ARRAY')
        layout.operator('group_pro.radial_array', icon='UNLINKED')


class GP_MT_SetOrigin(bpy.types.Menu):
    bl_idname = "GP_MT_set_origin"
    bl_label = "Set Group Origin"

    def draw(self, context):
        layout = self.layout
        op = layout.operator(GP_OT_GroupProSetOrginCenter.bl_idname, text="3D cursor", icon='PIVOT_CURSOR')
        op.method = '3DCURSOR'
        op = layout.operator(GP_OT_GroupProSetOrginCenter.bl_idname, text="Objects Center", icon='CENTER_ONLY')
        op.method = 'CENTER'
        op = layout.operator(GP_OT_GroupProSetOrginCenter.bl_idname, text="Bounding box bottom", icon='ALIGN_BOTTOM')
        op.method = 'BBB'
        op = layout.operator(GP_OT_GroupProSetOrginCenter.bl_idname, text="Bounding box center", icon='LIGHTPROBE_CUBEMAP')
        op.method = 'BBC'


class GP_MT_GroupProMenu(bpy.types.Menu):
    bl_label = "GroupPro"
    bl_idname = "GP_MT_gpro_menu"

    def draw(self, context):
        GP_PT_GroupProPanel.draw(self, context)
