#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####


import bpy
from pathlib import Path
from .update_addon import *
from .helpers_group import localView_up, addon_name_lowercase, get_addon_preferences, update_all_bb_objs, update_all_coll_colors
from bpy.props import FloatVectorProperty, FloatProperty, EnumProperty
from .gp_ui import GP_PT_GroupProPanel, VIEW3D_PT_Color
import rna_keymap_ui
##########################         Addon Prefernedces                ###############################################


MESSAGE = None
panels = (
    GP_PT_GroupProPanel,
    VIEW3D_PT_Color,
)


def update_panel(self, context):
    message = "GPro: Updating Panel locations has failed"
    try:
        for panel in panels:
            if "bl_rna" in panel.__dict__:
                bpy.utils.unregister_class(panel)

        for panel in panels:
            panel.bl_category = get_addon_preferences().category
            bpy.utils.register_class(panel)

    except Exception as e:
        print("\n[{}]\n{}\n\nError:\n{}".format(__name__, message, e))
        pass


class groupProPreferences(bpy.types.AddonPreferences):
    bl_idname = __package__

    def set_abs_path(self, context):
        abs_p = bpy.path.abspath(self['exchange_path'])
        global MESSAGE
        if os.path.exists(abs_p):
            try:
                f = open(abs_p+'write_test.txt', "w+")
            except OSError as e:
                print(e)
                self['exchange_path'] = str(Path.home())
                MESSAGE = f'No permission to write to:  {abs_p}. Select different folder'
            else:
                MESSAGE = None
                f.close()
                os.remove(abs_p+'write_test.txt')
                self['exchange_path'] = abs_p
        else:
            self['exchange_path'] = str(Path.home())

    def switch_bbox_verts(self, context):
        update_all_bb_objs()

    def update_colors(self, context):
        update_all_coll_colors()


    #* drawing props
    bbox_color: FloatVectorProperty(name="Bounding Box color", default=(0.0, 0.0, 0.0), size=3, subtype="COLOR", min=0, max=1)
    edited_box_color: FloatVectorProperty(name="Edited Group color", default=(0.0, 1.0, 0.0, 0.8), size=4, subtype="COLOR", min=0, max=1)
    box_line_width: FloatProperty(name="Edited Group Box line width", default=1, min=0, max=4, step=1, precision=3)
    groupBBox_use: bpy.props.BoolProperty(name="Draw Bounding Box around all Groups", description="Display Group Instances Bounding Box",  default=False)
    disable_bbox_verts: bpy.props.BoolProperty( name='Disable Bbox vertices', description='Disable Bounding Box vertices - you wont be able to snap to group cornerts', default=True, update=switch_bbox_verts)
    isolate_mode: bpy.props.EnumProperty(name="Isolate view mode", items=[("FLAT", "History", "Use edition history list for isolation of edited group objects"),
                                                                          ("TREE", "Collection depth", "Use collection nesting depth for isolation of edited objects")], default="TREE", update=localView_up)

    #* Linked lib props
    edited_lib_box_color: FloatVectorProperty(name="Edited Library color", default=(1.0, 0.4, 0.0, 0.8), size=4, subtype="COLOR", min=0, max=1)
    show_confirmation_linked_open: bpy.props.BoolProperty(name='Show confirmation on open', description='Show confirmation popup, when opening linked group (from external library)', default=True)
    exchange_path_type: bpy.props.EnumProperty(name="Exchange Folder Type", default="BLEND",
                                                description="Folder that will be used for storing temporary library data,\n"
                                                            "Libraries will read data from it",
                                                items=(("BLEND", "Use edited blend file location", "Write exchange data to same folder where currently opened blend file is"),
                                                       ("CUSTOM", "Define custom folder", "Pick one exchange folder - same for all blend files")))
    exchange_path: bpy.props.StringProperty(name="Exchange Path", default=str(Path.home()), description="Library exchange Path", subtype="DIR_PATH", update=set_abs_path)

    #* Others
    uninstance_meshes_in_gcopies: bpy.props.BoolProperty(name='Uninstance meshes on copy', description='Uninstance meshes when duplicatng group instances (shift + D)', default=True)
    uninstance_materials_in_gcopies: bpy.props.BoolProperty(name='Uninstance materials on copy', description='Uninstance meshes when duplicatng group instances (shift + D)', default=False)

    container_prefix_suffix: bpy.props.EnumProperty(name='', items=[('PREFIX', 'Prefix', ''), ('SUFFIX', 'Suffix', '')], default='PREFIX')
    group_prefix_suffix: bpy.props.EnumProperty(name='', items=[('PREFIX', 'Prefix', ''), ('SUFFIX', 'Suffix', '')], default='SUFFIX')

    container_prefix: bpy.props.StringProperty(name="Empty Prefix", description="Prefix added to empty when creating group",  default='gp_')
    group_prefix: bpy.props.StringProperty(name="Collection Prefix", description="Prefix added to group name on creation",  default='_coll')

    expandPanel: bpy.props.BoolProperty(name="", description="", default=False)
    quaternion_rot: bpy.props.BoolProperty(name="Use Quaternion Rotations", description="Euler rotation sometimes may lead to gimball lock. Quaternion rotations fixes that",  default=False)
    proxy_switch_on_render: bpy.props.BoolProperty(name='Proxy switch on render', description='', default=False)
    auto_link_children: bpy.props.BoolProperty(name="Auto link children", description="Move child objects into group automatically, when you create group from parent", default=False)
    auto_link_modifier_objs: bpy.props.BoolProperty(name="Auto link modifier objects",
                                                    description="Move modifier objects into group automatically, when you create group with modifier that uses that object",  default=True)
    #Colors
    auto_color: bpy.props.BoolProperty(name='Auto Viewport Color', description='Automatically assign random color to group objects (uses objects \'viewport display\' color property)', default=True, update=update_colors)



    reduction_method: bpy.props.EnumProperty(name='Reduction method', description='Smooth Axis',
                                             items=[('CONVEX', 'Convex Hull(Fastest)', 'Convex Hull - better Bounding Box'),
                                                    ('DECIMATE', 'Only Decimate', 'Use only Decimate modifier'),
                                                    ('REMESH', 'Remesh And Decim(Slow)', 'Use Remesh and decimate')
                                                    ], default='CONVEX')
    remesh_mode: bpy.props.EnumProperty(name="Remesh Smooth", description="", default="SMOOTH",
                                        items=(("SMOOTH", "Smooth", ""),
                                               ("SHARP", "Sharp", "")))
    octree_depth: bpy.props.IntProperty(name='octree_depth', default=6, soft_max=10, min=2)
    remove_disconnected: bpy.props.FloatProperty(name='Remove Disconnected threshold', default=0.1, min=0, max=1)
    decimate: bpy.props.FloatProperty(name='Decimate', default=0.8, min=0.01, max=1)



    tabs: bpy.props.EnumProperty(name="Tabs", items=[("UPDATE", "Update", ""),
                                                     ("HOTKEYS", "Hotkeys", ""),
                                                     ("NAMING", "Naming", ""),
                                                     ("DISPLAY", "Drawing", ""),
                                                     ("PROXY", "Proxies", ""),
                                                     ("LIBRARIES", "Libraries", ""),
                                                     # ("COLORS", "Colors", ""),
                                                     ("OTHER", "Extra", "")], default="UPDATE")
    update_exist: bpy.props.BoolProperty(name="Update Exist", description="There is new GroupPro update",  default=False)
    update_text: bpy.props.StringProperty(name="Update text",  default='')

    category: bpy.props.StringProperty(name="Tab Category", description="Choose a name for the category of the panel", default="Group Pro", update=update_panel)

    def draw(self, context):
        layout = self.layout
        row = layout.row(align=True)
        row.prop(self, "tabs", expand=True)
        box = layout.box()
        if self.tabs == "HOTKEYS":
            # wm = bpy.context.window_manager
            # kc = wm.keyconfigs.user  # cos we override addon hotkey,  with this
            # km = kc.keymaps['3D View']
            # kmi = get_hotkey_entry_item(km, 'wm.call_menu_pie', "GP_MT_grouppro_main_pie")
            # if kmi:
            #     col.context_pointer_set("keymap", km)
            #     rna_keymap_ui.draw_kmi([], kc, km, kmi, col, 0)
            # else:
            #     col.label(text="No hotkey entry found")
            #     col.operator(GP_OT_GroupPro_Add_Hotkey.bl_idname, text="Add hotkey entry", icon='PLUS')
            col = box.column()
            col.label(text="Keymap List:",icon="KEYINGSET")

            wm = context.window_manager
            old_km_name = ""
            found_addon_kmi = [] # list of tuples (exist, (km, kmi))
            for i,(km_add, kmi_add) in enumerate(addon_keymaps):
                for user_km in wm.keyconfigs.user.keymaps:
                    if km_add.name == user_km.name:
                        km = user_km
                        break
                got_kmi = False
                for kmi in km.keymap_items:
                    if kmi_add.idname == kmi.idname:
                        if kmi_add.name == kmi.name:
                            found_addon_kmi.append((True,(km,kmi)))
                            got_kmi = True
                if not got_kmi:
                    found_addon_kmi.append((False,(km_add,kmi_add)))


            # found_addon_kmi = sorted(set(found_addon_kmi), key=found_addon_kmi.index)

            kc = wm.keyconfigs.user
            for i, (exist, (km, kmi)) in enumerate(found_addon_kmi):
                if not km.name == old_km_name:
                    col.label(text=str(km.name),icon="DOT")
                if exist:   # user keymap
                    col.context_pointer_set("keymap", km)
                    rna_keymap_ui.draw_kmi([], kc, km, kmi, col, 0)
                    col.separator()
                else: # else holds addon keymap - that is missing in user keymap
                    split = col.split(factor=0.1)
                    split.label(text='   ')
                    op = split.operator(GP_OT_ReregisterUserKeym.bl_idname, text=f"Restore '{kmi.name}' hotkey" , icon="DISCLOSURE_TRI_RIGHT")
                    op.index = i
                    col.separator()

                old_km_name = km.name

            box.operator(GP_OT_GroupPro_Add_Hotkey.bl_idname, text="Reset Group Pro Hotkeys")

        if self.tabs == "UPDATE":
            col = box.column()
            sub_row = col.row(align=True)
            sub_row.operator(addon_name_lowercase()+".check_for_update")
            split_lines_text = self.update_text.splitlines()
            for line in split_lines_text:
                sub_row = col.row(align=True)
                sub_row.label(text=line)
            sub_row.separator()
            sub_row = col.row(align=True)
            if self.update_exist:
                sub_row.operator(addon_name_lowercase()+".update_addon", text='Install latest version').reinstall = False
            else:
                sub_row.operator(addon_name_lowercase()+".update_addon", text='Reinstall current version').reinstall = True
            sub_row.operator(addon_name_lowercase()+".rollback_addon")

        elif self.tabs == "NAMING":
            col = box.column()
            sub_row = col.row(align=True)
            sub_row.prop(self, "container_prefix")
            sub_row.prop(self, "container_prefix_suffix", text='')
            sub_row = col.row(align=True)
            sub_row.prop(self, "group_prefix")
            sub_row.prop(self, "group_prefix_suffix", text='')

        elif self.tabs == "LIBRARIES":
            col = box.column()
            sub_row = col.row(align=True)
            sub_row.prop(self, "edited_lib_box_color")
            sub_row = col.row(align=True)
            sub_row.prop(self, "exchange_path_type")
            if self.exchange_path_type == 'CUSTOM':
                sub_row = col.row(align=True)
                sub_row.prop(self, "exchange_path")
                sub_row = col.row(align=True)
                if MESSAGE:
                    sub_row.label(text=MESSAGE)
            sub_row = col.row(align=True)
            sub_row.prop(self, "show_confirmation_linked_open")

        elif self.tabs == "PROXY":
            col = box.column()
            col.prop(self, 'reduction_method')
            if self.reduction_method == 'REMESH':
                col.prop(self, 'remesh_mode')
                col.prop(self, 'octree_depth')
                col.prop(self, 'remove_disconnected')
            col.prop(self, 'decimate')


        elif self.tabs == "DISPLAY":
            row = box.row()
            box.label(text="Group Bounding Boxes properties")
            row = box.row(align=True)
            row.prop(self, "edited_box_color")
            row = box.row(align=True)
            row.prop(self, "groupBBox_use")
            if self.groupBBox_use:
                row = box.row(align=True)
                row.prop(self, "bbox_color")
            row = box.row(align=True)
            row.prop(self, "disable_bbox_verts")
            row = box.row(align=True)
            row.prop(self, "box_line_width")
            row = box.row(align=True)
            row.prop(self, "isolate_mode")
            row = box.row(align=True)
            row.prop(self, "auto_color")

        elif self.tabs == "OTHER":
            col = box.column(align=True)
            col.label(text="Tab Category:")
            col.prop(self, "category", text="")

            box.label(text="Mesh behaviour when duplicating group - Shift + D")
            row = box.row(align=True)
            row.prop(self, "uninstance_meshes_in_gcopies")
            if self.uninstance_meshes_in_gcopies:
                row.prop(self, "uninstance_materials_in_gcopies")

            row = box.row(align=True)
            box.label(text="Other")
            row = box.row(align=True)
            row.prop(self, "quaternion_rot")
            row.prop(self, "proxy_switch_on_render")

            row = box.row(align=True)
            row.label(text="Automatic linking of objects outside group:")
            row = box.row(align=True)
            row.prop(self, "auto_link_children")
            row.prop(self, "auto_link_modifier_objs")

#### end  MOve TO pref file ####################################


def get_hotkey_entry_item(km, kmi_name, kmi_value):
    '''
    returns hotkey of specific type, with specific properties.name (keymap is not a dict, so referencing by keys is not enough
    if there are multiple hotkeys!)
    '''
    for i, km_item in enumerate(km.keymap_items):
        if km.keymap_items.keys()[i] == kmi_name:
            if km.keymap_items[i].properties.name == kmi_value:
                return km_item
    return None



def new_keymap_items(kc):
    '''AFAIK this is overridden later by user set hotkey in keyconfigs.user.
    keyconfig.addon - is always there, but if entry is missing in keyconfig.user , then GroupPro hotkey won't work

    addon_km = wm.keyconfigs.addon   # always has gpro hotkey (and other addons)
    user_km = wm.keyconfigs.user # default (active) + addons. Final keyconfig, can be customized by user.  Fix: if gpro menu is missing here, ctrl+X wont work.
    active_km = wm.keyconfigs.active # same as default?
    default_km = wm.keyconfigs.default #same as active?
    '''

    new_keymaps = []

    # for kc in [addon_keyconfig, ]:
    # kc = addon_keyconfig
    km_3d = kc.keymaps.get('3D View')
    if not km_3d:
        km_3d = kc.keymaps.new(name='3D View', space_type="VIEW_3D")
    kmi = km_3d.keymap_items.new('wm.call_menu_pie', 'X', 'PRESS', ctrl=True)
    kmi.properties.name = "GP_MT_grouppro_main_pie"
    new_keymaps.append((km_3d, kmi))

    km_om = kc.keymaps.get('Object Mode')
    if not km_om:
        km_om = kc.keymaps.new(name='Object Mode', space_type="EMPTY")
    kmi2 = km_om.keymap_items.new('object.duplicate_group_override', 'D', 'PRESS', shift=True)
    kmi2.active = True
    new_keymaps.append((km_om, kmi2))

    kmi3 = km_3d.keymap_items.new('object.paste_group_override', 'V', 'PRESS', ctrl=True)
    kmi3.active = True
    new_keymaps.append((km_3d, kmi3))

    kmi3 = km_om.keymap_items.new("object.join_group_override", 'J', 'PRESS', ctrl=True)
    kmi3.active = True
    new_keymaps.append((km_om, kmi3))

    return new_keymaps

addon_keymaps = []

def register_keymap():
    wm = bpy.context.window_manager

    addon_keyconfig = wm.keyconfigs.addon   # always has gpro hotkey (and other addons)
    user_keyconfig = wm.keyconfigs.user # default (active) + addons. Final keyconfig, can be customized by user.  Fix: if gpro menu is missing here, ctrl+X wont work.
    active_keyconfig = wm.keyconfigs.active # same as default?
    default_km = wm.keyconfigs.default #same as active?

    registered_km_items = new_keymap_items(addon_keyconfig)
    global addon_keymaps
    for km, kmi in registered_km_items:
        addon_keymaps.append((km, kmi))

    km3 = wm.keyconfigs.default.keymaps['3D View']
    if 'view3d.pastebuffer' in km3.keymap_items.keys():  # won't be there on blender start ( is there better workaround?)
        km3.keymap_items['view3d.pastebuffer'].active = False

    km2 = wm.keyconfigs.default.keymaps['Object Mode']
    # kmi2 = km2.keymap_items.new('object.duplicate_group_override', 'D', 'PRESS', shift=True)
    if 'object.duplicate_move' in km2.keymap_items.keys():  # disable default hotkey if exist
        kmi2 = km2.keymap_items['object.duplicate_move'].active = False

    if 'object.join' in km2.keymap_items.keys():  # disable default hotkey if exist
        kmi2 = km2.keymap_items['object.join'].active = False


class GP_OT_GroupPro_Add_Hotkey(bpy.types.Operator):
    ''' Add hotkey entry '''
    bl_idname = "groupro.add_hotkey"
    bl_label = "GPro Init Hotkey"
    bl_options = {'REGISTER', 'INTERNAL'}

    def execute(self, context):
        wm = bpy.context.window_manager
        # wm.keyconfigs.user # addon + active (or default?) # is user removed Gpro ctrl+X key, then this will add it back
        registered_km_items = new_keymap_items(wm.keyconfigs.user)

        self.report({'INFO'}, "Restored Group Pro Hotkeys in User Preferences -> Input -> 3D view")
        return {'FINISHED'}


# operator GP_OT_ReregisterUserKeym, that will re-register user keymap from addon keymap item
class GP_OT_ReregisterUserKeym(bpy.types.Operator):
    bl_idname = "groupro.reregister_user_keym"
    bl_label = "Restore GroupPro Hotkey"
    bl_description = "Restore missing GroupPro hotkey entry"
    bl_options = {'REGISTER', 'INTERNAL'}

    index: bpy.props.IntProperty()

    def execute(self, context):
        (km_add,kmi_add) = addon_keymaps[self.index]
        wm = bpy.context.window_manager
        user_keyconfig = wm.keyconfigs.user
        km = user_keyconfig.keymaps.get(km_add.name)
        if not km:
            km = user_keyconfig.keymaps.new(name=km_add.name, space_type=km_add.space_type)
        kmi = km.keymap_items.new(kmi_add.idname, kmi_add.type, kmi_add.value, ctrl=kmi_add.ctrl, shift=kmi_add.shift, alt=kmi_add.alt, oskey=kmi_add.oskey)
        if hasattr(kmi_add.properties, 'name'):
            kmi.properties.name = kmi_add.properties.name
        self.report({'INFO'}, "Restored Group Pro Hotkeys in User Preferences -> Input -> 3D view")
        return {'FINISHED'}


def unregister_keymap():
    global addon_keymaps
    wm = bpy.context.window_manager
    for km, kmi in addon_keymaps:
        if len(km.keymap_items) > 0:  # why this is empty/broken of F8 scripts.reload?
            try:
                km.keymap_items.remove(kmi)
            except Exception as e:
                print(e)
        # if km in wm.keyconfigs.addon.keymaps.values():
        #     wm.keyconfigs.addon.keymaps.remove(km)
    addon_keymaps.clear()

    wm = bpy.context.window_manager
    km = wm.keyconfigs.default.keymaps['3D View']  # activate back defautl ctrl+v and shift+d hotkeys
    if 'view3d.pastebuffer' in km.keymap_items.keys():  # won't be there on blender start ( is there better workaround?)
        km.keymap_items['view3d.pastebuffer'].active = True

    km = wm.keyconfigs.default.keymaps['Object Mode']  # activate back defautl ctrl+v and shift+d hotkeys XXX: enablinig this break 3d view nav...
    if 'object.duplicate_move' in km.keymap_items.keys():  # won't be there on blender start ( is there better workaround?)
        km.keymap_items['object.duplicate_move'].active = True
    if 'object.join' in km.keymap_items.keys():  # won't be there on blender start ( is there better workaround?)
        km.keymap_items['object.join'].active = True
