'''
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import bpy
import sys
import os


# data_types = dir(bpy.data)
data_types = ['actions', 'armatures', 'cache_files', 'cameras', 'collections', 'curves', 'fonts', 'grease_pencils', 'images', 'lattices', 'lightprobes',
              'lights', 'linestyles', 'masks', 'materials', 'meshes', 'metaballs', 'movieclips', 'node_groups', 'objects', 'paint_curves', 'palettes', 'particles', 'scenes', 'sounds', 'speakers', 'texts', 'textures']


def append_element(import_blend_path, original_source_file):
    original_datas = {}
    for attr in data_types:
        original_datas[attr] = {data_id.name: data_id for data_id in getattr(bpy.data, attr)}
        # if original_datas[attr]:
        #     print(f'Cached: {attr}')


    # original_collections = {col.name: col for col in bpy.data.collections}
    original_source_path = os.path.dirname(original_source_file)
    current_file_path = bpy.context.blend_data.filepath
    current_file_dir = os.path.dirname(current_file_path)

    imported_datas_names = {}
    
    with bpy.data.libraries.load(import_blend_path, relative=False) as (data_from, data_to):
        # for attr in dir(data_to):
        for attr in data_types:
            setattr(data_to, attr, getattr(data_from, attr))
            imported_datas_names[attr] = [data_name for data_name in getattr(data_from, attr)]
            if imported_datas_names[attr]:
                print(f'Imported {attr}: imported_datas_names[attr]')
    os.remove(import_blend_path)

     #* it remaps objs, mat, text inside coll.objs too.
    # imported_col_names = imported_datas_names['collections']
    for attr, orig_data in original_datas.items():
        new_data_ids = getattr(data_to, attr)  # eg: data_to.collections
        data = getattr(bpy.data, attr)  # eg. bpy.data.collections
        for orig_name, old_data in orig_data.items(): #eg. for collections, objs, textures etc.
            if orig_name in imported_datas_names[attr]:
                new_data = new_data_ids[imported_datas_names[attr].index(orig_name)]
                if 'lib_path' in new_data:
                    del new_data['lib_path']
                if new_data.library: #if incoming id,  links to lib, then: reamap new_data.lib.path or skip (if point to current file)
                    lib_abs_path = new_data.library.filepath
                    
                    #! below wont happen (overriding circular ID's). Blender automatically skips imporitng ids linking to current file
                    # lib_path = new_data.library.filepath
                    # lib_abs_path = bpy.path.abspath(lib_path, start=original_source_path)
                    # if lib_abs_path == current_file_path: #importing id, with lib pointing to current file, so skip it
                    #     print(f'ID {new_data.name} lib points to current file. Skipping')
                    #     continue
                    
                    try:
                        lib_path = bpy.path.relpath(lib_abs_path, start=current_file_dir) #remap from  original file, to relative to current file
                    except:
                        lib_path = lib_abs_path
                    remaped_lib_path = lib_path
                    print(f'incoming id lib remapping from {lib_abs_path} to {remaped_lib_path} ')
                    new_data.library.filepath = remaped_lib_path
                        
                # print(f'remapping: {orig_name}:{new_data.name}')
                old_data.user_remap(new_data)
                new_data.name = orig_name
                data.remove(old_data)

    # #remap old coll, to imported from: data_to.collections[name] 
    # #* it remaps objs, mat, text inside coll.objs too.
    # imported_col_names = imported_datas_names['collections']
    # for orig_name, old_col in original_collections.items():
    #     if orig_name in imported_col_names:
    #         # for empty_dupli in old_col.users_dupli_group:
    #         #     empty_dupli.instance_collection = data_to.collections[imported_col_names.index(orig_name)]
    #         #remap original_collections to new imported onews
    #         old_col.user_remap(data_to.collections[imported_col_names.index(orig_name)])
    #         bpy.data.collections.remove(old_col)



if __name__ == '__main__':
    # folder = os.path.split(bpy.context.blend_data.filepath)[0]
    # print(sys.argv, len(sys.argv))
    # force_rerender = True if 'force_rerender' in sys.argv else False
    original_source_file = sys.argv[4]
    exhange_file = sys.argv[5]
    # exhange_file = "C://tmp//untitled.blend"
    print(f'Lib file({bpy.context.blend_data.filepath}):  Importing data from: {original_source_file}')
    append_element(exhange_file, original_source_file)

    bpy.ops.wm.save_mainfile()
    bpy.ops.wm.quit_blender()
