# ##### BEGIN GPL LICENSE BLOCK #####
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

import os
import bpy

ID_TO_PARENT = {'action': 'actions',
                'armature': 'armatures',
                'cache_file': 'cache_files',
                'camera': 'cameras',
                'collection': 'collections',
                'curve': 'curves',
                'font': 'fonts',
                'grease_pencil': 'grease_pencils',
                'image': 'images',
                'lattice': 'lattices',
                'lightprobe': 'lightprobes',
                'light': 'lights',
                'linestyle': 'linestyles',
                'mask': 'masks',
                'material': 'materials',
                'mesh': 'meshes',
                'metaball': 'metaballs',
                'movieclip': 'movieclips',
                'shader node tree': 'node_groups',
                'object': 'objects',
                'paint_curve': 'paint_curves',
                'palette': 'palettes',
                'particle': 'particles',
                'scene': 'scenes',
                'sound': 'sounds',
                'speaker': 'speakers',
                'text': 'texts',
                'texture': 'textures'}

BLENDER_ID_NAMES = [ 'collections', 'objects',
    'armatures', 'cameras', 'curves', 'meshes', 'metaballs', 'fonts', 'lightprobes', 'lattices', 'texts', 'lights',
    'materials', 'particles', 'images', 'textures',
    'node_groups']

LIB_EDIT_TIME = 0
INSTANCES_COL = {}

def delayed_load_lib(current_group, stored_lib_path):

    abs_lib_path = bpy.path.abspath(stored_lib_path)
    edit_time = os.path.getmtime(abs_lib_path)
    if edit_time <= LIB_EDIT_TIME:
        return 0.2  # wait every 0.2 sec,
    else:
        # lib = [lib for lib in bpy.data.libraries if lib.filepath == stored_lib_path][0]
        #get empties.dupli_ins that were pointng to lib
        linked_cols_local = [obj.instance_collection.name for obj in bpy.data.objects if obj.type ==
                             'EMPTY' and obj.instance_collection and 'lib_path' in obj.instance_collection and obj.instance_collection['lib_path'] == stored_lib_path]
        with bpy.data.libraries.load(abs_lib_path, link=True) as (data_from, data_to):
            # data_to.collections = [current_group.name] #? how de we get all id's that were imported with coll?
            # ? how de we get all id's that were imported with coll?
            data_to.collections = [name for name in data_from.collections if name in linked_cols_local]

        #? What if we have obj, with same name as in lib? IT will be overriden.. unless
        #? maybe tag obj - from_lib, when opening group_lib ?
        if data_to.collections:
            for new_id in data_to.collections[0].library.users_id:  # eg. obj1, obj2,  col1, mat1 etc
                id_type = ID_TO_PARENT[str.lower(new_id.bl_rna.name)]  # 'objects', 'collections' etc
                existing_ids = getattr(bpy.data, id_type)  # eg. bpy.data.objects
                for old_id in existing_ids:
                    if old_id.name == new_id.name and old_id != new_id:
                        old_id.user_remap(new_id)
                        existing_ids.remove(old_id)  # eg. bpy.data.objects.remove
        return None  # unregister timer




def make_local_all(lib):
    ''' make lib.users_id local (unlike: bpy.ops.object.make_local(type='ALL'), which affects all bpy.data.libs) '''
    
    def localize_names(link_id):
        #* rename local datablocks to prevent library ids name change on make_local() 
        id_type = ID_TO_PARENT[str.lower(link_id.bl_rna.name)]  # 'objects', 'collections' etc
        existing_ids = [id for id in getattr(bpy.data, id_type) if id != link_id]  # eg. collections #? exclude other libs?
        for existing_id in existing_ids:
            if existing_id.name == link_id.name:
                i = 1
                existing_id.name += '.001'
                while existing_id.name in existing_ids:
                    existing_id.name = existing_id.name[:-3] + str(i).zfill(3)
                    i += 1
        #mark sub groups as from lib too. bpy.ops.object.make_local() - erases obj lib info
        if link_id.bl_rna.name == 'Object' and link_id.type == 'EMPTY' and link_id.instance_collection:
            link_id.instance_collection['lib_path'] = lib.filepath
        link_id.make_local()  # ! main thing


    #* start from root lib collection. Find it first.
    lib_cols = [el for el in lib.users_id if el.bl_rna.name == 'Collection']
    root_cols = pick_root_cools(lib_cols)

    global INSTANCES_COL
    INSTANCES_COL.clear()
    for r_coll in root_cols:
        make_coll_local_recursive(r_coll)

    #*order matters when makin local. First parents
    for bl_attr in BLENDER_ID_NAMES[2:]: #skip colls, and objs as they are done above
        for bl_id in getattr(bpy.data, bl_attr): #objects, colls etc
            if bl_id.library == lib:
                localize_names(bl_id)

def pick_root_cools(colls):
    ''' chec if each collection in cools, has any parent? If not return collection as root '''
    if not colls:
        return []
    root_cools = set()
    lib = colls[0].library
    for searched_col in colls:
        is_root = True #at begnining assume searched coll is root. Then try to falsify it.
        empty_dupli_users = [obj for obj in searched_col.users_dupli_group if obj.library and obj.library == lib]
        if empty_dupli_users:  # scan scan if any obj.users_collection - if not then searched_col is root.
            for dupli_obj in empty_dupli_users:
                parent = [d_col for d_col in dupli_obj.users_collection if d_col.library and d_col.library == lib]
                if parent:  # searched col, has one empty - dupli_obj, which is used by d_col from lib, so no is_root
                    is_root = False
                    break
        else:
            pass #may be still root if has parent_cools
        if is_root:
            parent_cols = [coll for coll in colls if searched_col.name in coll.children.keys()]
            if parent_cols:
                is_root = False
                break
        if is_root:
            root_cools.add(searched_col)
    return list(root_cools)


def make_coll_local_recursive(col):
    ''' make coll and col.objects local.
        Its there, to make sure, that all sub_objs, are not linked to edited_lib, to avoid circular depenency, on import to lib.
        Skips imgs, particles etc. : ( - this will give circular dep - handeld in make_local_all
    '''
    edited_lib = col.library
    global INSTANCES_COL
    def inner_make_local(col):
        if col.library:
            if col.library == edited_lib:
                if col.name in INSTANCES_COL:
                    local_col = INSTANCES_COL[col.name]
                else:
                    local_col = col.make_local()
                    local_col['lib_path'] = edited_lib.filepath
                    INSTANCES_COL[local_col.name] = local_col
            else:  # empty.inst_coll links to other lib, so just skipp checking anyting
                return
        else:  # empty.inst_coll is not linked. But countinuech checking children - to remove linnks to edited_lib
            local_col = col

        for obj in local_col.all_objects:
            obj.make_local()
            if obj.type == 'EMPTY' and obj.instance_collection:
                inner_make_local(obj.instance_collection)

    inner_make_local(col)



#OLD
def make_local_sel(empty):
    ''' make local selected empty, and empty.instace_coll, objs, data.
        Make sure all sub_objs, are not linked to edited_lib, to avoid circular depenency.
        Skips imgs, particles etc. : ( - this will give circular dep
    '''
    instances_data = {}
    instances_mat = {}
    instances_col = {}

    #! what about imgs, particles etc.
    edited_lib = None
    if empty.library:
        edited_lib = empty.library
    elif empty.instance_collection.library:
        edited_lib = empty.instance_collection.library
    if not edited_lib:
        print(f'Empty {empty.name} is not in library. Cant be made local')
        return empty
    # else:
    #     lib_path = empty.instance_collection['lib_path']
    #     libs = [l for l in bpy.data.libraries if l.filepath == lib_path]
    #     lib = libs[0] if libs else None
    def inner_make_local(empty):
        if empty.library:
            if empty.library == edited_lib:
                empty_local = empty.make_local()
            else: #empt is from not edited likned lib, so just skip it and contained objs (if any)
                return empty
        else:
            empty_local = empty

        group = empty_local.instance_collection
        if group.library:
            if group.library == edited_lib:
                if group.name in instances_col:
                    local_group = instances_col[group.name]
                else:
                    local_group = group.make_local()
                    local_group['lib_path'] = edited_lib.filepath
                    instances_col[local_group.name] = local_group
            else: #empty.inst_coll links to other lib, so just skipp checking anyting
                return
        else:  # empty.inst_coll is not linked. But countinuech checking children - to remove linnks to edited_lib
            local_group = group

        for obj in local_group.all_objects:
            if obj.type == 'EMPTY' and obj.instance_collection:
                local_obj = inner_make_local(obj)
            else:
                local_obj = obj.make_local()
            if obj.data:  # make local data
                if obj.data.name in instances_data:
                    local_obj.data = instances_data[obj.data.name]
                    local_data = local_obj.data
                else:  # add to instances list
                    local_data = local_obj.data.make_local()
                    instances_data[obj.data.name] = local_data
                    for i, mat in enumerate(local_data.materials):
                        if mat.name in instances_mat:
                            local_data.materials[i] = instances_mat[mat.name]
                        else:
                            local_mat = mat.make_local()
                            instances_mat[mat.name] = local_mat

        return empty_local


    local_empty = inner_make_local(empty)
    return local_empty
