#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ##### END GPL LICENSE BLOCK #####

import bpy

class Node(object):
    def __init__(self, data):
        self.data = data
        self.children = {}

    def add_child(self, elem):
        self.children[elem] = elem  # cos faster than lists 10x


class TreeData(object):
    masterRoot = None

    def __init__(self):
        self.masterRoot = {}

    def clear(self):
        self.masterRoot.clear()

    def __SearchAndInsertRec(self, parentNode, valueToFind, elementToInsert):
        if parentNode.data == valueToFind:
            parentNode.add_child(Node(elementToInsert))
            return True
        for childNode in parentNode.children.keys():
            out = self.__SearchAndInsertRec(childNode, valueToFind, elementToInsert)
            if out:
                return True
        return False

    # return True if element was found
    def SearchAndInsert(self, valueToFind, elementToInsert):
        for node in self.masterRoot.keys():
            if self.__SearchAndInsertRec(node, valueToFind, elementToInsert):
                return True
        self.masterRoot[Node(elementToInsert)] = 1  # if element was not found, then just add node to root

    # return depth of element if was

    def __FindElementDepthRec(self, parentBranch, valueToFind, depth=1):
        if parentBranch.data == valueToFind:
            return depth
        for childNode in parentBranch.children.keys():
            res = self.__FindElementDepthRec(childNode, valueToFind, depth + 1)
            if res > 0:
                return res
        return 0

    # return depth of element if was
    def FindElementDepth(self, valueToFind, depth=0):
        for node in self.masterRoot.keys():
            dep = self.__FindElementDepthRec(node, valueToFind)
            if dep > 0:
                return dep
        return 0

    def __GetMaxDepthRec(self, parentBranch, depth=1):
        if len(parentBranch.children) == 0:
            return depth
        maxDepth = 0
        for child in parentBranch.children.keys():
            maxDepth = max(self.__GetMaxDepthRec(child, depth + 1), maxDepth)
        return maxDepth

    # return depth of element if was
    def GetMaxDepth(self):
        depth = 0
        for node in self.masterRoot.keys():
            depth = max(self.__GetMaxDepthRec(node), depth)
        return depth

    def __PrintValuesRec(self, parentNode):
        output = '[' + parentNode.data + ']'
        if len(parentNode.children) == 0:
            return output
        for childNode in parentNode.children.keys():
            output += '->[' + self.__PrintValuesRec(childNode) + ']'
        return output

    def __repr__(self):
        output = ''
        for node in self.masterRoot.keys():
            output += self.__PrintValuesRec(node) + '\n'
        return output

